# -*- coding: utf-8 -*-

"""This module contains all the constants used in Bio2BEL MSIGDB"""

import os

from bio2bel.utils import get_connection, get_data_dir

MODULE_NAME = 'msig'
DATA_DIR = get_data_dir(MODULE_NAME)
DEFAULT_CACHE_CONNECTION = get_connection(MODULE_NAME)

CONFIG_FILE_PATH = os.path.join(DATA_DIR, 'config.ini')

PATHWAY_LINK = 'http://www.broadinstitute.org/gsea/msigdb/cards/{}'

RESOURCES_PATH = os.path.abspath(os.path.join(__file__, "../../.."))

GENE_SETS_PATH = os.path.join(DATA_DIR, 'resources', 'gene_sets.gmt')



# -*- coding: utf-8 -*-

"""
Bio2BEL MSig is a package for enriching BEL networks with MsigDB. Furthermore, it is integrated in the `ComPath environment <https://github.com/ComPath>`_ for pathway database comparison.

Citation
--------
- Subramanian, A., Tamayo, P., Mootha, V. K., Mukherjee, S., Ebert, B. L., Gillette, M. A., ... & Mesirov, J. P. (2005). Gene set enrichment analysis: a knowledge-based approach for interpreting genome-wide expression profiles. Proceedings of the National Academy of Sciences, 102(43), 15545-15550.
- Liberzon, A., Subramanian, A., Pinchback, R., Thorvaldsdóttir, H., Tamayo, P., & Mesirov, J. P. (2011). Molecular signatures database (MSigDB) 3.0. Bioinformatics, 27(12), 1739-1740.


"""

from .manager import Manager

__version__ = '0.0.5'

__title__ = 'bio2bel_msig'
__description__ = "A wrapper around MSIGDB"
__url__ = 'https://github.com/bio2bel/MSIG'

__author__ = 'Daniel Domingo-Fernández and Charles Tapley Hoyt'
__email__ = 'daniel.domingo.fernandez@scai.fraunhofer.de'

__license__ = 'MIT License'
__copyright__ = 'Copyright (c) 2017-2018 Daniel Domingo-Fernández and Charles Tapley Hoyt'



# -*- coding: utf-8 -*-

from .constants import GENE_SETS_PATH


def _process_line(line):
    """Returns pathway name, url and gene sets associated

    :param str line: gmt file line
    :rtype: str
    :return: pathway name
    :rtype: str
    :return: pathway info url
    :rtype: list[str]
    :return: genes set associated
    """
    processed_line = [
        word.strip()
        for word in line.split('\t')
    ]

    return processed_line[0], processed_line[1], processed_line[2:]


def parse_gmt_file(url=None):
    """Returns file as list of pathway - gene sets

    :param Optional[str] url: url from gmt file
    :return: line-based processed file
    :rtype: list
    """
    pathways = []
    with open(url if url else GENE_SETS_PATH, 'r') as file:
        for line in file:
            pathway_name, url_info, gene_set = _process_line(line)

            pathways.append((pathway_name, url_info, gene_set))

    return pathways



# -*- coding: utf-8 -*-
""" This module contains all test constants"""

import os
import tempfile
import unittest

from bio2bel_msig.manager import Manager

dir_path = os.path.dirname(os.path.realpath(__file__))
resources_path = os.path.join(dir_path, 'resources')

gene_sets_path = os.path.join(resources_path, 'test_gmt_file.gmt')


class DatabaseMixin(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Create temporary file"""

        cls.fd, cls.path = tempfile.mkstemp()
        cls.connection = 'sqlite:///' + cls.path

        # create temporary database
        cls.manager = Manager(cls.connection)

        # fill temporary database with test data
        cls.manager.populate(
            path=gene_sets_path
        )

    @classmethod
    def tearDownClass(cls):
        """Closes the connection in the manager and deletes the temporary database"""
        cls.manager.drop_all()
        cls.manager.session.close()
        os.close(cls.fd)
        os.remove(cls.path)



# -*- coding: utf-8 -*-

"""This module populates the tables of Bio2BEL MSIG."""

import logging
import os

from compath_utils import CompathManager
from tqdm import tqdm

from .constants import MODULE_NAME
from .models import Base, Pathway, Protein
from .parser import parse_gmt_file

__all__ = [
    'Manager'
]

log = logging.getLogger(__name__)


class Manager(CompathManager):
    """Bio2BEL MSIG manager."""

    module_name = MODULE_NAME

    flask_admin_models = [Pathway, Protein]
    pathway_model = Pathway
    protein_model = Protein
    pathway_model_identifier_column = Pathway.msig_id

    @property
    def _base(self):
        return Base

    def summarize(self):
        """Summarize the database.

        :rtype: dict[str,int]
        """
        return dict(
            pathways=self._count_model(Pathway),
            proteins=self._count_model(Protein),
        )

    """Custom query methods"""

    def get_or_create_pathway(self, pathway_name):
        """Gets an pathway from the database or creates it

        :param str pathway_name: pathway ame
        :rtype: Pathway
        """
        pathway = self.get_pathway_by_name(pathway_name)

        if pathway is None:
            pathway = Pathway(
                name=pathway_name,
                msig_id=pathway_name
            )
            self.session.add(pathway)

        return pathway

    def get_protein_by_id(self, identifier):
        """Gets a protein by its id

        :param identifier: identifier
        :rtype: Optional[Protein]
        """
        return self.session.query(Protein).filter(Protein.id == identifier).one_or_none()

    def get_or_create_protein(self, hgnc_symbol):
        """Gets an protein from the database or creates it
        :param Optional[str] hgnc_symbol: name of the protein
        :rtype: Protein
        """
        protein = self.get_protein_by_hgnc_symbol(hgnc_symbol)

        if protein is None:
            protein = Protein(
                hgnc_symbol=hgnc_symbol,
            )
            self.session.add(protein)

        return protein

    def get_protein_by_hgnc_symbol(self, hgnc_symbol):
        """Gets a protein by its hgnc symbol

        :param hgnc_id: hgnc identifier
        :rtype: Optional[Protein]
        """
        return self.session.query(Protein).filter(Protein.hgnc_symbol == hgnc_symbol).one_or_none()

    """Methods to populate the DB"""

    def populate(self, path=None):
        """Populates all tables

        :param Optional[str] path: url from gmt file
        """
        if path is None:
            path = os.environ.get('BIO2BEL_MSIG_PATH')

            if path is None:
                raise RuntimeError('No path set in the environment. Please specify the path of the file to populate')

        pathways = parse_gmt_file(url=path)

        hgnc_symbol_protein = {}

        for pathway_name, _, gene_set in tqdm(pathways, desc='Loading database'):

            pathway = self.get_or_create_pathway(pathway_name=pathway_name)

            for hgnc_symbol in gene_set:

                protein = self.get_or_create_protein(hgnc_symbol)
                hgnc_symbol_protein[hgnc_symbol] = protein

                if pathway not in protein.pathways:
                    protein.pathways.append(pathway)

            self.session.commit()

    def _add_admin(self, app, **kwargs):
        from flask_admin import Admin
        from flask_admin.contrib.sqla import ModelView

        class PathwayView(ModelView):
            """Pathway view in Flask-admin"""
            column_searchable_list = (
                Pathway.id,
                Pathway.name
            )

        class ProteinView(ModelView):
            """Protein view in Flask-admin"""
            column_searchable_list = (
                Protein.hgnc_symbol,
                Protein.id
            )

        admin = Admin(app, **kwargs)
        admin.add_view(PathwayView(Pathway, self.session))
        admin.add_view(ProteinView(Protein, self.session))
        return admin



# -*- coding: utf-8 -*-

""" This module contains the flask application to visualize the db"""

from .manager import Manager

if __name__ == '__main__':
    manager = Manager()
    app = manager.get_flask_admin_app()
    app.run(debug=True, host='0.0.0.0', port=5000)



# -*- coding: utf-8 -*-
""" This module contains tests for parsing GMT files"""

from bio2bel_msig.models import Pathway, Protein

from tests.constants import DatabaseMixin


class TestParse(DatabaseMixin):
    """Tests the parsing module"""

    def test_pathway_count(self):
        pathway_number = self.manager.session.query(Pathway).count()
        self.assertEqual(3, pathway_number)

    def test_protein_count(self):
        protein_number = self.manager.session.query(Protein).count()
        self.assertEqual(13, protein_number)

    def test_pathway_protein_1(self):
        pathway = self.manager.get_pathway_by_id('AAANWWTGC_UNKNOWN')
        self.assertIsNotNone(pathway, msg='Unable to find pathway')
        self.assertEqual(4, len(pathway.proteins))

        self.assertEqual(
            {'MEF2C', 'ATP1B1', 'RORA', 'PDS5B'},
            {
                gene.hgnc_symbol
                for gene in pathway.proteins
            }
        )

    def test_pathway_protein_2(self):
        pathway = self.manager.get_pathway_by_id('AAAYRNCTG_UNKNOWN')
        self.assertIsNotNone(pathway, msg='Unable to find pathway')
        self.assertEqual(3, len(pathway.proteins))

        self.assertEqual(
            {'LTBP1', 'LEKHM1', 'PDS5B'},
            {
                gene.hgnc_symbol
                for gene in pathway.proteins
            }
        )

    def test_gene_query_1(self):
        """Single protein query. This protein is associated with 1 pathways"""
        enriched_pathways = self.manager.query_gene_set(['KCNE1L'])
        self.assertIsNotNone(enriched_pathways, msg='Enriching function is not working')

        self.assertEqual(
            {
                "pathway_id": "MYOD_01",
                "pathway_name": "MYOD_01",
                "mapped_proteins": 1,
                "pathway_size": 8,
                "pathway_gene_set": {
                    'DST',
                    'EFNA1',
                    'EIF2C1',
                    'FAM126A',
                    'HMGN2',
                    'KCNE1L',
                    'PDS5B',
                    'PGF'
                }
            },
            enriched_pathways["MYOD_01"]
        )

    def test_gene_query_2(self):
        """Multiple protein query"""
        enriched_pathways = self.manager.query_gene_set(['PDS5B', 'ATP1B1'])
        self.assertIsNotNone(enriched_pathways, msg='Enriching function is not working')

        self.assertEqual(
            {
                "pathway_id": "AAANWWTGC_UNKNOWN",
                "pathway_name": "AAANWWTGC_UNKNOWN",
                "mapped_proteins": 2,
                "pathway_size": 4,
                "pathway_gene_set": {'PDS5B', 'ATP1B1', 'MEF2C', 'RORA'}
            },

            enriched_pathways["AAANWWTGC_UNKNOWN"]
        )

        self.assertEqual(
            {
                "pathway_id": "AAAYRNCTG_UNKNOWN",
                "pathway_name": "AAAYRNCTG_UNKNOWN",
                "mapped_proteins": 1,
                "pathway_size": 3,
                "pathway_gene_set": {'PDS5B', 'LEKHM1', 'LTBP1'}
            },
            enriched_pathways["AAAYRNCTG_UNKNOWN"]
        )

        self.assertEqual(
            {
                "pathway_id": "MYOD_01",
                "pathway_name": "MYOD_01",
                "mapped_proteins": 1,
                "pathway_size": 8,
                "pathway_gene_set": {
                    'DST',
                    'EFNA1',
                    'EIF2C1',
                    'FAM126A',
                    'HMGN2',
                    'KCNE1L',
                    'PDS5B',
                    'PGF'
                }
            },
            enriched_pathways["MYOD_01"]
        )



# -*- coding: utf-8 -*-

"""Bio2BEL MSIG database models"""

from sqlalchemy import Column, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()

TABLE_PREFIX = 'msig'
PATHWAY_TABLE_NAME = '{}_pathway'.format(TABLE_PREFIX)
PROTEIN_TABLE_NAME = '{}_protein'.format(TABLE_PREFIX)
PROTEIN_PATHWAY_TABLE = '{}_protein_pathway'.format(TABLE_PREFIX)

protein_pathway = Table(
    PROTEIN_PATHWAY_TABLE,
    Base.metadata,
    Column('protein_id', Integer, ForeignKey('{}.id'.format(PROTEIN_TABLE_NAME)), primary_key=True),
    Column('pathway_id', Integer, ForeignKey('{}.id'.format(PATHWAY_TABLE_NAME)), primary_key=True)
)


class Pathway(Base):
    """Pathway Table"""

    __tablename__ = PATHWAY_TABLE_NAME

    id = Column(Integer, primary_key=True)

    msig_id = Column(String(255), unique=True, index=True, nullable=False, doc='id')

    name = Column(String(255), unique=True, index=True, nullable=False, doc='pathway name')

    proteins = relationship(
        'Protein',
        secondary=protein_pathway,
        backref='pathways'
    )

    def __repr__(self):
        return self.name

    def get_gene_set(self):
        """Returns the genes associated with the pathway (gene set). Note this function restricts to HGNC symbols genes

        :rtype: set[bio2bel_msig.models.Protein]
        """
        return {
            protein.hgnc_symbol
            for protein in self.proteins
            if protein.hgnc_symbol
        }

    @property
    def resource_id(self):
        return self.msig_id

    @property
    def url(self):
        return 'http://software.broadinstitute.org/gsea/msigdb/geneset_page.jsp?geneSetName={}'.format(self.name)


class Protein(Base):
    """Genes Table"""

    __tablename__ = PROTEIN_TABLE_NAME

    id = Column(Integer, primary_key=True)

    hgnc_symbol = Column(String(255), unique=True, index=True, nullable=False, doc='hgnc symbol of the protein')

    def __repr__(self):
        return self.hgnc_symbol

    def get_pathways_ids(self):
        """Returns the pathways associated with the protein"""
        return {
            pathway.msig_id
            for pathway in self.pathways
        }
